/*
* thread.c
*
* Created: 17.04.2017 12:45:32
* Author: Detlev Tietjen
* Licence: LGPL 3.0
*/

#include <stddef.h>
#include <stdbool.h>
#include <avr/interrupt.h>
#include "thread.h"

#define THREAD_DIVISOR ((F_CPU * THREAD_MS_PER_TICK) / (1000UL * 1024UL))
#if (65535 < THREAD_DIVISOR)
#error THREAD_MS_PER_TICK to high
#endif

struct threadTCBData_t
{
	uint8_t state;
	bool	alreadyRunning;
	threadFunc_t func;
	void * argument;
	uint16_t period;
	uint16_t elapsed;
};

static uint8_t nextThreadIndex = 0;
static struct threadTCBData_t thread[THREAD_MAX];

pThread_t threadAdd(threadFunc_t f, void * argument, uint16_t periodTicks)
{
	pThread_t rp = NULL;
	
	if((0 < periodTicks) && (THREAD_MAX > nextThreadIndex))
	{
		rp = thread + nextThreadIndex++;
		rp->state			= 0;
		rp->alreadyRunning	= false;
		rp->func			= f;
		rp->argument		= argument;
		rp->period			= periodTicks;
		rp->elapsed			= periodTicks; //start initialization immediately
	}
	return rp;
}


void threadStart()
{
	TCC1.PER		= THREAD_DIVISOR;
	TCC1.CNT		= THREAD_DIVISOR - 1;
	TCC1.INTCTRLA	= TC_OVFINTLVL_LO_gc;
	TCC1.CTRLA		= TC_CLKSEL_DIV1024_gc;
	PMIC.CTRL |= PMIC_LOLVLEN_bm | PMIC_MEDLVLEN_bm | PMIC_HILVLEN_bm;
	sei();
}

/*
222:	0b b6       	in	r0, 0x3b	; 59
224:	0f 92       	push	r0
226:	af 92       	push	r10
228:	bf 92       	push	r11
22a:	cf 92       	push	r12
22c:	df 92       	push	r13
22e:	ef 92       	push	r14
230:	ff 92       	push	r15
232:	0f 93       	push	r16
234:	1f 93       	push	r17
236:	2f 93       	push	r18
238:	3f 93       	push	r19
23a:	4f 93       	push	r20
23c:	5f 93       	push	r21
23e:	6f 93       	push	r22
240:	7f 93       	push	r23
242:	8f 93       	push	r24
244:	9f 93       	push	r25
246:	af 93       	push	r26
248:	bf 93       	push	r27
24a:	cf 93       	push	r28
24c:	df 93       	push	r29
24e:	ef 93       	push	r30
250:	ff 93       	push	r31
*/
ISR(TCC1_OVF_vect, ISR_NAKED)
{
	asm volatile(
	"rcall reenableISR\r\n"
	"push r0\r\n"
	"in r0,__SREG__\r\n"
	"push r0\r\n"
	"push r1\r\n"
	"clr r1\r\n"
	#ifdef __AVR_HAVE_RAMPZ__
	"in		r0,__RAMPZ__\r\n"
	"push	r0\r\n"
	#endif
	"push	r10\r\n"
	"push	r11\r\n"
	"push	r12\r\n"
	"push	r13\r\n"
	"push	r14\r\n"
	"push	r15\r\n"
	"push	r16\r\n"
	"push	r17\r\n"
	"push	r18\r\n"
	"push	r19\r\n"
	"push	r20\r\n"
	"push	r21\r\n"
	"push	r22\r\n"
	"push	r23\r\n"
	"push	r24\r\n"
	"push	r25\r\n"
	"push	r26\r\n"
	"push	r27\r\n"
	"push	r28\r\n"
	"push	r29\r\n"
	"push	r30\r\n"
	"push	r31\r\n");
	
	bool		higherThreadAlreadyRunning		= false;
	pThread_t	pThread							= thread;
	uint8_t		threadCnt						= nextThreadIndex; //total count of active tasks
	
	while(threadCnt--)
	{
		if(!higherThreadAlreadyRunning)
		{
			if(pThread->alreadyRunning)
			{
				higherThreadAlreadyRunning = true;
			}
			else
			{
				threadBeginCriticalBlock();
				if(pThread->elapsed >= pThread->period)
				{
					pThread->elapsed -= pThread->period;
					pThread->alreadyRunning = true;
				}
				threadEndCriticalBlock();
				
				if(pThread->alreadyRunning) // This could only be true if set in the block before
				{
					pThread->state = pThread->func(pThread->state, pThread->argument); //call the thread
					pThread->alreadyRunning = false;
				}
			}
		}
		threadBeginCriticalBlock();
		if(UINT16_MAX > pThread->elapsed) pThread->elapsed++;
		threadEndCriticalBlock();
		pThread++;
	}
	
	asm volatile(
	"pop	r31\r\n"
	"pop	r30\r\n"
	"pop	r29\r\n"
	"pop	r28\r\n"
	"pop	r27\r\n"
	"pop	r26\r\n"
	"pop	r25\r\n"
	"pop	r24\r\n"
	"pop	r23\r\n"
	"pop	r22\r\n"
	"pop	r21\r\n"
	"pop	r20\r\n"
	"pop	r19\r\n"
	"pop	r18\r\n"
	"pop	r17\r\n"
	"pop	r16\r\n"
	"pop	r15\r\n"
	"pop	r14\r\n"
	"pop	r13\r\n"
	"pop	r12\r\n"
	"pop	r11\r\n"
	"pop	r10\r\n"
	#ifdef __AVR_HAVE_RAMPZ__
	"pop	r0\r\n"
	"out	__RAMPZ__,r0\r\n"
	#endif
	"pop	r1\r\n"
	"pop	r0\r\n"
	"out	__SREG__,r0\r\n"
	"pop	r0\r\n"
	"ret\r\n"
	"reenableISR: reti\r\n");
	/*
	"push r0\r\n"
	"in r0,__SREG__\r\n"
	"push r0\r\n"
	"push r1\r\n"
	"clr r1\r\n"
	#ifdef __AVR_HAVE_RAMPZ__
	"in		r0,__RAMPZ__\r\n"
	"push	r0\r\n"
	#endif
	*/
}